/******************************************************************************
	Gaussian Elimination with Backward Substitution - Algorithm 6.1
*******************************************************************************

To solve the n x n linear system

	E1:   a(1,1)*X1 + a(1,2)*X2 + ... a(1,n)*Xn = a(1,n+1)
	E2:   a(2,1)*X1 + a(2,2)*X2 + ... a(2,n)*Xn = a(2,n+1)
	.        .           .               .           .
	.        .           .               .           .
	.        .           .               .           .
	En:   a(n,1)*X1 + a(n,2)*X2 + ... a(n,n)*Xn = a(n,n+1):

INPUT number of equations and unknowns n; augmented matrix A = (a(i,j)) where
1 <= i <= n and 1 <= j <= n+1.

OUTPUT solution x1,x2,...,xn or message that linear system has no unique
solution.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */


main()
{
  float **A, **E, **m, *X, sum, temp;
  int i, j, k, n, p, FLAG;

  get_title();

  printf("Gaussian Elimination with Backward Substitution - Algorithm 6.1\n\n");
  printf("Solves A*X=B.\n\n");
  fprintf(stderr,
      "Gaussian Elimination with Backward Substitution - Algorithm 6.1\n\n");
  fprintf(stderr, "Solves A*X=B.\n\n");

  fprintf(stderr, "Enter the dimension, n, of the matrix A: ");
  scanf("%d", &n);
  if (n <= 0)
    naaerror("n must be > 0.");	/* ERROR. */

  E = matrix(1,n,1,n+1);	/* Matrix A augmented with B. */
  m = matrix(1,n,1,n+1);	/* Matrix m. */
  X = vector(1,n);		/* Solution vector. */
  A = E;			/* Matrix A (a subset of E). */

  fprintf(stderr, "Enter the coefficients for matrix A:\n");
  printf("A = ");
  for (i=1;i<=n;i++) {
    printf("[ ");
    for (j=1;j<=n;j++) {
      fprintf(stderr, "\tA[%d][%d] = ", i, j);
      scanf("%f", &A[i][j]);
      printf("% 3g ", A[i][j]);
    }
    printf("]\n    ");
  }
  printf("\n");

  fprintf(stderr, "Enter the coefficients for vector B:\n");
  printf("B = [ ");
  for (i=1;i<=n;i++) {
    fprintf(stderr, "\tB[%d] = ", i);
    scanf("%f", &E[i][n+1]);
    printf("% 3g ", E[i][n+1]);
  }
  printf("]\n\n");

  /* STEP #1 */
  for (i=1;i<n;i++) {		/* Do Steps 2-4.  Elimination process. */

    printf("At row %d:\n", i);
    printf("E = ");
    print_matrix(E,n);		/* Show intermediate results for looks. */

    /* STEP #2 */
    for (p=i;p<=n;p++) {	/* Let p be the smallest integer with */
      if (A[p][i] != 0.0) {	/* i <= p <= n and a(pi) != 0. */
        FLAG = 1;
        break;
      }
      FLAG = 0;			/* FLAG = 0 indicates an all zero row. */
    }

    if (i == p)
      printf("No rows need swapping.\n\n");	/* For looks. */
    else
      printf("Need to swap row %d with row %d.\n\n", i, p);

    if (FLAG == 0) {
      printf("No unique solution exists (1).\n");
      exit (-1);		/* STOP */
    }

    /* STEP #3 */
    if (p != i)			/* Swap Ep with Ei. */
      for (k=1;k<=n+1;k++) {
        temp = E[p][k];
        E[p][k] = E[i][k];
        E[i][k] = temp;
      }

    /* STEP #4 */
    for (j=i+1;j<=n;j++) {	/* Do Steps 5-6. */

      /* STEP #5 */
      m[j][i] = A[j][i]/A[i][i];

      /* STEP #6 */
      for (k=1;k<=n+1;k++)
        E[j][k] -= m[j][i]*E[i][k];
    }
  }

  printf("At row %d:\n", n);
  printf("E = ");
  print_matrix(E,n);		/* Show intermediate results for looks. */

  /* STEP #7 */
  if (A[n][n] == 0.0) {
    printf("No unique solution exists (2).\n");
    exit (-1);			/* STOP */
  }

  /* STEP #8 */
  X[n] = E[n][n+1] / A[n][n];	/* Start backward substitution. */

  /* STEP #9 */
  for (i=n-1;i>=1;i--) {
    sum = 0.0;
    for (j=i+1;j<=n;j++)
      sum += A[i][j]*X[j];
    X[i] = (E[i][n+1] - sum) / A[i][i];
  }

  /* STEP #10 */
  printf("\nX = [ ");		/* Procedure completed successfully. */
  for (i=1;i<=n;i++)
    printf("% .8g ", X[i]);
  printf("]\n");

}  /* STOP */

/*****************************************************************************/
print_matrix(E,n)	/* Prints out a matrix. */
float **E;
int n;
{
  int i, j;

  for (i=1;i<=n;i++) {
    printf("[ ");
    for (j=1;j<=n;j++)
      printf("% 3g ", E[i][j]);
    printf(" | % 3g ", E[i][n+1]);
    printf("]\n    ");
  }
  printf("\n");

}
/*****************************************************************************/
