/******************************************************************************
		LDLt Factorization - Algorithm 6.5
*******************************************************************************

To factor the positive definite n x n matrix A into the form LDLt, where L is
a lower triangular matrix with ones along the diagonal and D is a diagonal
matrix with positive entries on the diagonal:

INPUT dimension n; entries (a(i,j)) 1 <= i,j <= n.

OUTPUT the entries l(i)(j), 1 <= j <= i, 1 <= i <= n of L and d(i), 1 <= i <= n of D.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */


main()
{
  double **A, **L, *D, *V, sum;
  int i, j, k, n;

  get_title();

  printf("LDLt Factorization - Algorithm 6.5\n\n");
  fprintf(stderr, "LDLt Factorization - Algorithm 6.5\n\n");

  fprintf(stderr, "Enter the dimension of matrix A: ");
  scanf("%d", &n);
  if (n <= 0)
    naaerror("n must be > 0.");	/* ERROR. */

  A = dmatrix(1,n,1,n);		/* Matrix A. */
  L = dmatrix(1,n,1,n);		/* Matrix L. */
  D = dvector(1,n);		/* Vector D. */
  V = dvector(1,n);		/* Vector V. */

  fprintf(stderr, "Enter the coefficients for matrix A:\n");
  printf("A = ");
  for (i=1;i<=n;i++) {
    printf("[ ");
    for (j=1;j<=n;j++) {
      fprintf(stderr, "\tA[%d][%d] = ", i, j);
      scanf("%lf", &A[i][j]);
      printf("% 9.9lg ", A[i][j]);
    }
    printf("]\n    ");
  }
  printf("\n\n");

  /* Remove the below comments to factor a symmetric matrix. */
  /*
  for (i=1;i<=n;i++)
    if (A[i][i] == 0.0)
      naaerror("Can't be done.  There is a zero diagonal element.");
  */

  for (i=1;i<=n;i++)		/* Put 1's on diagonal of L. */
    L[i][i] = 1;		/* Should be in text! */

  /* STEP #1 */
  for (i=1;i<=n;i++) {		/* Do Steps 2-4. */

    /* STEP #2 */
    for (j=1;j<i;j++)
      V[j] = L[i][j] * D[j];	/* The algorithm in the text is incorrect. */
				/* Replace A[i][j] with L[i][j]. */
    /* STEP #3 */
    sum = 0.0;
    for (j=1;j<i;j++)		/* The algorithm in the text is incorrect. */
      sum += L[i][j] * V[j];	/* Replace A[i][j] with L[i][j]. */

    D[i] = A[i][i] - sum;

    /* STEP #4 */
    for (j=i+1;j<=n;j++) {
      sum = 0.0;
      for (k=1;k<i;k++)		/* The algorithm in the text is incorrect. */
        sum += L[j][k] * V[k];	/* Replace A[j][k] with L[j][k]. */
      L[j][i] = (A[j][i] - sum) / D[i];
    }
  }

  /* STEP #5 */
  printf("L  = ");		/* Print out L. */
  print_matrix(L, n);

  for (i=1;i<=n;i++)		/* Put vector D into a matrix, A. */
    for (j=1;j<=n;j++)
      if (i == j)
        A[i][j] = D[i];
      else
        A[i][j] = 0.0;

  printf("D  = ");		/* Print out D. */
  print_matrix(A, n);

  for (i=1;i<=n;i++)		/* Make Lt, storing it in A. */
    for (j=1;j<=n;j++)
      A[i][j] = L[j][i];

  printf("Lt = ");		/* Print out Lt. */
  print_matrix(A, n);

}  /* STOP */

/*****************************************************************************/
print_matrix(A,n)	/* Prints out a matrix. */
double **A;
int n;
{
  int i, j;

  for (i=1;i<=n;i++) {
    printf("[ ");
    for (j=1;j<=n;j++)
      printf("% 9.9lg ", A[i][j]);
    printf("]\n     ");
  }
  printf("\n");

}
/*****************************************************************************/

