/******************************************************************************
	Least-Squares Polynomial Approximation - Algorithm 8.0B
*******************************************************************************

To determine a Least-Squares polynomial of degree n given m data points or the
function f(x).  Uses NORMAL EQUATIONS on page 431 for discrete data.  Uses
NORMAL EQUATIONS on page 440 for f(x).  Uses Simpson's rule to approximate
the integrals needed for continuous functions.

INPUT Degree of polynomial, n; number of data points, m; function f() if
desired; point to evaluate P(x) at, x.

OUTPUT Coefficients of an nth degree polynomial and optional values of P(x)
evaluated from the polynomial at x.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* For scanf() and printf() */
#include <math.h>		/* for math functions */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */
#include "gaussj.h"		/* To solve a matrix. */
#define	PI	3.14159265358979323846264338327950288419716939937510582097


main()
{
  double **A, *B, *X, *Y, **AA, f(), simpson(), x;
  int i, j, k, n, m, ch;

  get_title();

  printf("Least-Squares Polynomial Approximation - Algorithm 8.0B\n\n");
  fprintf(stderr,
	"Least-Squares Polynomial Approximation - Algorithm 8.0B\n\n");

  fprintf(stderr, "Enter the degree of the polynomial, n: ");
  scanf("%d", &n);
  if (n < 0)
    naaerror("n must be > or = to 0.");	/* ERROR. */

  AA = dmatrix(1,n+1,1,n+1);	/* Summation matrix */
  A = dmatrix(0,n+1,1,1);	/* Polynomial coefficients */

  fprintf(stderr,
	"Will you evaluate a function, f(x), or tabulated data? (f/d): ");
  ch = getchar();		/* Filters out Carriage Return/Line Feed. */
  ch = tolower(getchar());

  if (ch == 'd') {		/* For Tabulated Data. */
    fprintf(stderr, "How many points need evaluating?: ");
    scanf("%d", &m);
    if (m <= 0)
      naaerror("m must be > 0.");	/* ERROR. */
    if (n >= m-1)
      naaerror("Not enough data points for a polynomial of that degree.");

    X = dvector(1,m);		/* X input */
    Y = dvector(1,m);		/* Y input or f(x) points. */

    /* Get (x,y) points. */
    for (i=1;i<=m;i++) {	/* Enter X(i). */
      fprintf(stderr, "\nEnter x%d: ", i);
      scanf("%lf", &X[i]);
      fprintf(stderr, "Enter y%d: ", i);
      scanf("%lf", &Y[i]);
    }

    /* Output (x,y) points. */
    printf("i\t x\t y\n");
    printf("--------------------------\n");
    for (i=1;i<=m;i++)
      printf("%d\t %lG\t% lG\n", i, X[i], Y[i]);
    printf("\n");

    /* Build matrix elements for AX=B type solution. */
    for (i=0;i<=n;i++) {
      for (j=0;j<=n;j++) {
        AA[i+1][j+1] = 0.0;
        for (k=1;k<=m;k++)
          if ((i+j) != 0)
            AA[i+1][j+1] += pow(X[k], (double) (i+j));
          else
            AA[i+1][j+1]++;
      }
      A[i+1][1] = 0.0;
      for (k=1;k<=m;k++)
        if (i != 0)
          A[i+1][1] += Y[k] * pow(X[k], (double) (i));
        else
          A[i+1][1] += Y[k];
    }

  } else if (ch == 'f') {		/* For Functions. */

    double a, b;			/* Interval values. */

    printf("f(x) = sin(PI*x)\n\n");
    /* Needs updating    $    */

    fprintf(stderr, "Enter lower interval value, a: ");
    scanf("%lf", &a);

    fprintf(stderr, "Enter upper interval value, b: ");
    scanf("%lf", &b);

    fprintf(stderr, "Computing integrals with Simpson's Rule");

    /* Build matrix elements for AX=B type solution. */
    for (i=0;i<=n;i++) {
      for (j=0;j<=n;j++) {
        AA[i+1][j+1] = simpson(1,i+j,a,b);	/* For x^(i+j) */
      }
      A[i+1][1] = simpson(2,i,a,b);		/* For x^i*f(x) */
    }
    fprintf(stderr, "\n");

  } else			/* Neither 'd' or 'f'. */
    naaerror("Invalid choice.  Choose 'f' for f(x) or 'd' for data.");

  printf("The polynomial coefficient matrix is:\n\n");
  for (i=1;i<=n+1;i++) {
    printf("[ ");
    for (j=1;j<=n+1;j++)
      printf("% 9lg ", AA[i][j]);
    printf("][ a%d ] = [ %9lg ]\n", i-1, A[i][1]);
  }
  printf("\n");


  gaussj(AA,n+1,A,1);		/* Solve the matrix on p. 431 */


  for (i=0;i<=n;i++)		/* Want A[0..n] not A[1..n+1]. */
    A[i][1] = A[i+1][1];

  /* Output polynomial coefficients. */
  print_poly(A,n);		/* Show results in polynomial format. */

  printf("\n");
  fprintf(stderr, "Would you like to evaluate P(x)? (y/n): ");
  ch = getchar();		/* filters out New Line/Carrage Return */
  ch = tolower(getchar());

  while (ch == 'y') {
    double tmp;

    fprintf(stderr, "\nEnter x: ");
    scanf("%lf", &x);

    /* P(x) = Summation of  a(i)*x^i  summed from i = 0 to n. */
    if (n == 0)
      tmp = A[n][1];
    else {
      tmp = A[n][1]*x;
      for (i=n-1;i>0;i--)
        tmp = (tmp + A[i][1]) * x;
      tmp += A[0][1];
    }

    fprintf(stderr, "P(%lg) = %.9lg\n", x, tmp);
    printf("P(%lg) = %.9lg\n", x, tmp);

    fprintf(stderr, "\nWould you like to evaluate P(x) again? (y/n): ");
    ch = getchar();		/* filters out New Line/Carrage Return */
    ch = tolower(getchar());
  }

}  /* STOP */

/******************************************************************************
		Composite Simpson's Rule - Algorithm 4.1 Modified
******************************************************************************/
double simpson(type, j, a, b)
int type, j;
double a, b;
{
  double h, X, XI, XI0, XI1, XI2, f();
  int i, n = 100;

  fprintf(stderr, ".");		/* For feedback since it takes so long. */

  /* STEP #1 */
  h = (b-a)/n;

  /* STEP #2 */
  if (type == 1) {
    if (j != 0)
      XI0 = pow(a,(double)j) + pow(b,(double)j);
    else
      return (b-a);
  } else {	/* type == 2 */
    if (j != 0)
      XI0 = pow(a,(double)j)*f(a) + pow(b,(double)j)*f(b);
    else
      XI0 = f(a) + f(b);
  }

  XI1 = 0.0;			/* Summation of f(x(2i-1)). */
  XI2 = 0.0;			/* Summation of f(x(2i)). */

  /* STEP #3 */
  for (i=1;i<n;i++) {

    /* STEP #4 */
    X = a + i*h;

    /* STEP #5 */
    if (i % 2 == 0) {		/* For i even. */      
      if (type == 1) {
        if (j != 0)
          XI2 += pow(X,(double)j);
        else
          XI2 += f(X);
      } else {		/* type == 2 */
        if (j != 0)
          XI2 += pow(X,(double)j) * f(X);
        else
          XI2 += f(X);
      }

    } else {			/* For i odd. */
      if (type == 1) {
        if (j != 0)
          XI1 += pow(X,(double)j);
        else
          XI1 += f(X);
      } else {		/* type == 2 */
        if (j != 0)
          XI1 += pow(X,(double)j) * f(X);
        else
          XI1 += f(X);
      }
    }
  }

  /* STEP #6 */
  XI = h*(XI0 + 2.0*XI2 + 4.0*XI1) / 3.0;

  /* STEP #7 */
  return (XI);
}  				/* STOP */

/*****************************************************************************/
print_poly(a, n)	/* Prints out a Polynomial, with real coefficients, */
double **a;		/* in a pleasant format. */
int n;			/* Example: P(x) = -X^4 - 2*X^2 + 3*X + 7 */
{
  int i;

  printf("P(x) = ");
  for (i=n;i>=1;i--) { 		/* Special case: a(n), ..., a(1). */
    if (a[i][1] != 0.0) {	/* Don't print if coefficient is zero. */
      if (a[i][1] > 0.0) {	/* positive real part. */
        if (a[i][1] == 1.0) {	/* Special case: coefficient = 1. */
          if (i == 1)  printf("X");  /* a(1) */
          else  printf("X^%d", i);
        } else {
          if (i == 1)  printf("%g*X", a[i][1]);  /* a(1) */
          else  printf("%g*X^%d", a[i][1], i);
        }
      } else {			/* negative real part. */
        if (a[i][1] == -1.0) {	/* Special case: coefficient = -1. */
          if ((i == n) && (i != 1))  printf("-X^%d", i);  /* a(n) */
          else if ((i == n) && (i == 1))  printf("-X", i);  /* a(n) */
          else if (i == 1)  printf("X");  /* a(1) */
          else  printf("X^%d", i);
        } else {				       /* a(n) */
          if ((i == n) && (i != 1))  printf("-%g*X^%d", -a[i][1], i);
          else if (i == 1)  printf("%g*X", -a[i][1]);  /* a(1) */
          else  printf("%g*X^%d", -a[i][1], i);
        }
      }
    }
    if (a[i-1][1] > 0.0)     /* look at next coefficient to determine sign. */
      printf(" + ");
    if (a[i-1][1] < 0.0)
      printf(" - ");
  }

  if (a[0][1] > 0.0)		/* special case:  a(0). */
      printf("%g\n", a[0][1]);
  if (a[0][1] < 0.0)
    printf("%g\n", -a[0][1]);
}
/*****************************************************************************/
/*****************************************************************************/
double f(x)		/* Function to evaluate, f(x).  May not be needed. */
double x;
{
  return(sin(PI*x));
}
/*****************************************************************************/
