/******************************************************************************
			Power Method - Algorithm 9.1
*******************************************************************************

To approximate the dominant eigenvalue and associated eigenvector of the n x n
matrix A given a nonzero vector X:

INPUT dimension n; matrix A; vector X; tolerance TOL; maximum number of
iterations N.

OUTPUT approximate eigenvalue mu; approximate eigenvector X (with
infinite_norm(X) = 1) or a message that the maximum number of iterations was
exceeded.

*******************************************************************************
*	Written by:  Harold A. Toomey, BYU, Summer 1989, v4.11		      *
******************************************************************************/

#include <stdio.h>		/* for scanf() and printf() */
#include <math.h>		/* for fabs() */
#include "naa_util.h"		/* Numerical Analysis Algorithms Utilities. */

int n;				/* Number of equations and unknowns. */


main()
{
  double **A, *X, *Y, *TEMP, TOL;
  double ERR, mu, xp, yp, infinite_norm();
  int i, j, k, N;

  get_title();

  printf("Power Method - Algorithm 9.1\n\n");
  fprintf(stderr, "Power Method - Algorithm 9.1\n\n");

  fprintf(stderr, "Enter maximum number of iterations, N: ");
  scanf("%d", &N);
  printf("Maximum number of iterations: %d\n", N);

  fprintf(stderr, "Enter the tolerance, TOL (1.0e-4): ");
  scanf("%lf", &TOL);
  printf("Tolerance = %lg\n", TOL);

  fprintf(stderr, "Enter the dimension, n, of the matrix A: ");
  scanf("%d", &n);
  if (n <= 0)
    naaerror("n must be > 0.");	/* ERROR. */
  printf("System has %d equations and %d unknowns.\n\n", n, n);

  A = dmatrix(1,n,1,n);		/* Matrix A */
  X = dvector(1,n);		/* solution vector */
  Y = dvector(1,n);		/* Vector Y */
  TEMP = dvector(1,n);		/* Temporary matrix for error */

  fprintf(stderr, "Enter the coefficients for matrix A:\n");
  printf("A = ");
  for (i=1;i<=n;i++) {
    printf("[ ");
    for (j=1;j<=n;j++) {
      fprintf(stderr, "\tA[%d][%d] = ", i, j);
      scanf("%lf", &A[i][j]);
      if (A[i][j] >= 0.0)  printf(" ");
      printf("%lg  ", A[i][j]);
    }
    printf("]\n    ");
  }
  printf("\n");

  fprintf(stderr, "Enter initial Eigenvector, X:\n");
  printf("X = [  ");
  for (i=1;i<=n;i++) {
    fprintf(stderr, "\tX[%d] = ", i);
    scanf("%lf", &X[i]);
    printf("%lg  ", X[i]);
  }
  printf("]t\n\n");

  /*-------------------------------------------------------------------------*/
  printf("\n k\t mu = mu\t\t");		/* print out table header */
  for (i=1;i<=n;i++)
    printf(" X[%d]\t\t", i);
  printf("\n----------");
  for (i=1;i<=n+1;i++)
    printf("---------------");
  printf("\n");
  /*-------------------------------------------------------------------------*/

  printf("% d\t\t\t", 0);	/* Print first row of output table. */
  for (i=1;i<=n;i++)
    printf("% .9lf\t", X[i]);
  printf("\n");

  /* STEP #1 */
  k = 1;

  /* STEP #2 */
  xp = infinite_norm(X);

  /* STEP #3 */
  for (i=1;i<=n;i++)
    X[i] /= xp;

  /* STEP #4 */
  while (k <= N) {		/* Do Steps 5-11. */

    /* STEP #5 */
    for (i=1;i<=n;i++) {
      Y[i] = 0.0;
      for (j=1;j<=n;j++)
        Y[i] += A[i][j] * X[j];
    }

    /* STEPS #6 & #7 */		/* Error: Text should swap these steps. */
    yp = infinite_norm(Y);
    mu = yp;

    /* STEP #8 */
    if (yp == 0.0) {
      printf("\n\neigenvector is:\nX = [");
      for (i=1;i<=n;i++)
        printf("\t %.11lf", X[i]);
      printf(" ]\n\nA has eigenvalue 0, select new vector X and restart.\n");
      exit (-1);		/* STOP */
    }

    /* STEP #9 */
    for (i=1;i<=n;i++)
      TEMP[i] = X[i] - Y[i] / yp;
    ERR = fabs(infinite_norm(TEMP));
    for (i=1;i<=n;i++)
      X[i] = Y[i] / yp;

    /* STEP #10 */
    printf("% d\t", k);
    printf("% .9lf\t", mu);
    for (i=1;i<=n;i++)
      printf("% .9lf\t", X[i]);
    printf("\n");

    if (ERR < TOL)
      exit (1);			/* STOP - Procedure completed successfully. */

    /* STEP #11 */
    k++;

  }

  /* STEP #12 */		/* Procudure completed unseccessfully. */
  printf("\nMaximum number of iterations (%d) exceeded.\n", N);

}  /* STOP */

/****************************************************************************/
double infinite_norm(X)
double *X;
{
  double max = 0.0;
  int i;

  for (i=1;i<=n;i++)
    if (fabs(X[i]) > fabs(max))
      max = X[i];
  return (max);
}
/****************************************************************************/
